import functools
import json
from dataclasses import dataclass
from enum import StrEnum
from typing import Any, TypeVar

from aiohttp import MultipartReader, BodyPartReader
from aiohttp.web_exceptions import HTTPBadRequest
from multidict import CIMultiDictProxy, CIMultiDict

from .mime_types import ContentType


class MultipartHeaders(StrEnum):
    content_type = 'Content-Type'


T = TypeVar("T")


@dataclass(frozen=True)
class BodyPart:
    filename: str | None
    header_name_to_value: CIMultiDictProxy[str]
    content: bytes
    encoding: str | None

    def __post_init__(self):
        assert (content_type := self.content_type) is None or isinstance(content_type, str)

    @functools.cached_property
    def content_type(self) -> str | None:
        content_type = None
        content_types: list[str] | None = self.get_all_header_values(MultipartHeaders.content_type)
        if content_types:
            if len(content_types) > 1:
                raise HTTPBadRequest(reason=str(f'Body part has more than one content type: {content_types}'))
            content_type = content_types[0]
        return content_type

    @functools.cached_property
    def size(self) -> int:
        return len(self.content)

    def parse(self, default_content_type: ContentType = ContentType.Text) -> Any:
        content_type = self.content_type or default_content_type
        if content_type == ContentType.Text:
            return self.content.decode(self.encoding or "utf-8")
        elif content_type == ContentType.Json:
            return json.loads(self.content.decode(self.encoding or "utf-8"))
        else:
            return self.content

    def get_one_header_value(self, name: str, default: T = None) -> str | T:
        return self.header_name_to_value.getone(name, default)

    def get_all_header_values(self, name: str, default: T = None) -> list[str] | T:
        return self.header_name_to_value.getall(name, default)

    def __str__(self):
        parts = []
        parts += [f"filename={self.filename}"] if self.filename is not None else []
        parts += [f"content_type={self.content_type}"] if self.content_type is not None else []
        if self.header_name_to_value.keys() - {MultipartHeaders.content_type}:
            headers = [
                repr(f'{k}:{v}')
                for k, v in self.header_name_to_value.items()
                if k != MultipartHeaders.content_type
            ]
            parts += [f"headers={{{','.join(headers)}}}"]
        parts += f'size={self.size}'
        return f'BodyPart({", ".join(parts)})'


async def parse_multipart_form_data(reader: MultipartReader) -> CIMultiDictProxy[BodyPart]:
    result: CIMultiDict[BodyPart] = CIMultiDict()

    async for part in reader:
        part: BodyPartReader
        # noinspection PyTypeChecker
        encoding: str | None = part.get_charset(default=None)
        if part.name is None:
            raise HTTPBadRequest(reason="Malformed body part name")
        result.add(part.name, BodyPart(part.filename, part.headers, await part.read(decode=True), encoding=encoding))

    return CIMultiDictProxy(result)



from enum import unique, StrEnum


@unique
class ContentType(StrEnum):
    Atom = 'application/atom+xml'
    EDI_X12 = 'application/EDI-X12'
    EDIFACT = 'application/EDIFACT'
    Yaml = 'application/yaml'
    OldYaml = 'application/x-yaml'  # deprecated from 2024
    TextYaml = 'text/yaml'  # deprecated from 2024
    Json = 'application/json'
    TextJson = 'text/json'
    ProblemJson = 'application/problem+json'
    VndJson = 'application/vnd.api+json'
    HalJson = 'application/hal+json'
    JavaScript = 'application/javascript'
    Octet = 'application/octet-stream'
    Ogg = 'application/ogg'
    PDF = 'application/pdf'
    PostScript = 'application/postscript'
    SOAP = 'application/soap+xml'
    Web = 'application/font-woff'
    XHTML = 'application/xhtml+xml'
    DTD = 'application/xml-dtd'
    XOP = 'application/xop+xml'
    ZIP = 'application/zip'
    Gzip = 'application/gzip'
    BitTorrent = 'application/x-bittorrent'
    TeX = 'application/x-tex '
    XML = 'application/xml'
    DOC = 'application/msword'
    FormUrlEncoded = 'application/x-www-form-urlencoded'
    DVI = 'application/x-dvi'
    LaTeX = 'application/x-latex'
    TTF = 'application/x-font-ttf'
    AdobeFlash = 'application/x-shockwave-flash'
    StuffIt = 'application/x-stuffit'
    RAR = 'application/x-rar-compressed'
    Tarball = 'application/x-tar'
    XJavascript = 'application/x-javascript'
    MicrosoftExcel = 'application/vnd.ms-excel'
    MicrosoftPowerpoint = 'application/vnd.ms-powerpoint'
    XLSX = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    Mozilla = 'application/vnd.mozilla.xul+xml'

    Mulaw = 'audio/basic'
    PCM24bit = 'audio/L24'
    Mp4Audio = 'audio/mp4'
    AAC = 'audio/aac'
    MP3 = 'audio/mpeg'
    OggVorbis = 'audio/ogg'
    Vorbis = 'audio/vorbis'
    WindowsAudio = 'audio/x-ms-wma'
    WindowsMedia = 'audio/x-ms-wax'
    RealAudio = 'audio/vnd.rn-realaudio'
    WAV = 'audio/vnd.wave'
    WebMAudio = 'audio/webm'

    GIF = 'image/gif'
    JPEG = 'image/jpeg'
    PJPEG = 'image/pjpeg'
    PNG = 'image/png'
    SVG = 'image/svg+xml'
    TIFF = 'image/tiff'
    ICO = 'image/vnd.microsoft.icon'
    WBMP = 'image/vnd.wap.wbmp'
    WebP = 'image/webp'
    RGB = 'image/x-rgb'
    BMP = 'image/bmp'
    MS_BMP = 'image/x-ms-bmp'
    WINDOWS_BMP = 'image/x-windows-bmp'

    HTTP = 'message/http'
    IMDN = 'message/imdn+xml'
    EmailPartial = 'message/partial'
    Email = 'message/rfc822'

    ModelExample = 'model/example'
    IGS = 'model/iges'
    MSH = 'model/mesh'
    WRL = 'model/vrml'
    X3DB = 'model/x3d+binary'
    X3DV_VRML = 'model/x3d+vrml'
    X3D_XML = 'model/x3d+xml'

    MultipartMixed = 'multipart/mixed'
    MultipartAlternative = 'multipart/alternative'
    MultipartRelated = 'multipart/related'
    MultipartForm = 'multipart/form-data'
    MultipartSigned = 'multipart/signed'
    MultipartEncrypted = 'multipart/encrypted'

    ICalendar = 'text/calendar'
    CMD = 'text/cmd'
    Css = 'text/css'
    CSV = 'text/csv'
    HTML = 'text/html'
    Text = 'text/plain'
    PHP = 'text/php'
    XmlText = 'text/xml'
    Markdown = 'text/markdown'
    Cache = 'text/cache-manifest'
    jQuery = 'text/x-jquery-tmpl'
    TSV = 'text/tab-separated-values'

    MPEG = 'video/mpeg'
    MKV = 'video/x-matroska'
    Mp4Video = 'video/mp4'
    OggVideo = 'video/ogg'
    QuickTime = 'video/quicktime'
    WebMVideo = 'video/webm'
    WindowsVideo = 'video/x-ms-wmv'
    FLV = 'video/x-flv'
    gpp = 'video/3gpp'
    gpp2 = 'video/3gpp2'
    TS = 'video/mp2t'
    M3U8 = 'video/x-mpegURL'
    AVI = 'video/x-msvideo'
